import equinox as eqx
import jax
import jax.numpy as jnp
import pytest

from jaxonlayers.functions.masking import make_causal_mask
from jaxonlayers.layers import (
    Transformer,
    TransformerDecoder,
    TransformerDecoderLayer,
    TransformerEncoder,
    TransformerEncoderLayer,
)


class TestTransformerEncoderLayer:
    @pytest.mark.parametrize(
        "is_causal, use_explicit_mask",
        [
            (False, False),
            (True, False),
            (False, True),
            (True, True),
        ],
    )
    def test_masking(self, is_causal, use_explicit_mask):
        d_model = 64
        nhead = 4
        seq_len = 10

        layer = TransformerEncoderLayer(
            d_model=d_model,
            nhead=nhead,
            key=jax.random.key(0),
            dtype=jnp.float32,
        )

        x = jax.random.normal(jax.random.key(1), (seq_len, d_model))
        mask = make_causal_mask(seq_len) if use_explicit_mask else None
        output = layer(x, attn_mask=mask, is_causal=is_causal, inference=True)

        assert output.shape == (seq_len, d_model)

    def test_jit_no_retrace(self):
        d_model = 64
        nhead = 4
        seq_len = 10

        layer = TransformerEncoderLayer(
            d_model=d_model,
            nhead=nhead,
            key=jax.random.key(0),
            dtype=jnp.float32,
        )

        @eqx.filter_jit
        @eqx.debug.assert_max_traces(max_traces=1)
        def forward(model, x, mask, is_causal, key, inference):
            return model(x, mask, is_causal, key, inference=inference)

        x = jax.random.normal(jax.random.key(1), (seq_len, d_model))

        output1 = forward(layer, x, None, False, None, True)
        output2 = forward(layer, x, None, False, None, True)
        output3 = forward(layer, x, None, False, None, True)

        assert output1.shape == (seq_len, d_model)
        assert output2.shape == (seq_len, d_model)
        assert output3.shape == (seq_len, d_model)


class TestTransformerDecoderLayer:
    @pytest.mark.parametrize(
        "tgt_is_causal, memory_is_causal, use_tgt_mask, use_memory_mask",
        [
            (False, False, False, False),
            (True, False, False, False),
            (False, False, True, False),
            (False, False, False, True),
            (True, False, True, True),
        ],
    )
    def test_masking(
        self, tgt_is_causal, memory_is_causal, use_tgt_mask, use_memory_mask
    ):
        d_model = 64
        nhead = 4
        tgt_len = 10
        src_len = 12

        layer = TransformerDecoderLayer(
            d_model=d_model,
            nhead=nhead,
            key=jax.random.key(0),
            dtype=jnp.float32,
        )

        tgt = jax.random.normal(jax.random.key(1), (tgt_len, d_model))
        memory = jax.random.normal(jax.random.key(2), (src_len, d_model))
        tgt_mask = make_causal_mask(tgt_len) if use_tgt_mask else None
        memory_mask = (
            jnp.ones((tgt_len, src_len), dtype=jnp.bool_) if use_memory_mask else None
        )

        output = layer(
            tgt,
            memory,
            tgt_mask=tgt_mask,
            memory_mask=memory_mask,
            tgt_is_causal=tgt_is_causal,
            memory_is_causal=memory_is_causal,
            inference=True,
        )

        assert output.shape == (tgt_len, d_model)

    def test_with_process_heads(self):
        d_model = 64
        nhead = 4
        tgt_len = 10
        src_len = 12

        layer = TransformerDecoderLayer(
            d_model=d_model,
            nhead=nhead,
            key=jax.random.key(0),
            dtype=jnp.float32,
        )

        def identity_process_heads(q, k, v):
            return q, k, v

        tgt = jax.random.normal(jax.random.key(1), (tgt_len, d_model))
        memory = jax.random.normal(jax.random.key(2), (src_len, d_model))
        output = layer(
            tgt,
            memory,
            process_heads_sa=identity_process_heads,
            process_heads_mha=identity_process_heads,
            inference=True,
        )

        assert output.shape == (tgt_len, d_model)

    def test_jit_no_retrace(self):
        d_model = 64
        nhead = 4
        tgt_len = 10
        src_len = 12

        layer = TransformerDecoderLayer(
            d_model=d_model,
            nhead=nhead,
            key=jax.random.key(0),
            dtype=jnp.float32,
        )

        @eqx.filter_jit
        @eqx.debug.assert_max_traces(max_traces=1)
        def forward(model, tgt, memory, inference):
            return model(tgt, memory, inference=inference)

        tgt = jax.random.normal(jax.random.key(1), (tgt_len, d_model))
        memory = jax.random.normal(jax.random.key(2), (src_len, d_model))

        output1 = forward(layer, tgt, memory, True)
        output2 = forward(layer, tgt, memory, True)
        output3 = forward(layer, tgt, memory, True)

        assert output1.shape == (tgt_len, d_model)
        assert output2.shape == (tgt_len, d_model)
        assert output3.shape == (tgt_len, d_model)


class TestTransformerEncoder:
    @pytest.mark.parametrize(
        "is_causal, use_explicit_mask",
        [
            (False, False),
            (True, False),
            (False, True),
            (True, True),
        ],
    )
    def test_masking(self, is_causal, use_explicit_mask):
        d_model = 64
        nhead = 4
        num_layers = 3
        seq_len = 10

        encoder = TransformerEncoder(
            d_model=d_model,
            nhead=nhead,
            num_layers=num_layers,
            key=jax.random.key(0),
            dtype=jnp.float32,
        )

        x = jax.random.normal(jax.random.key(1), (seq_len, d_model))
        mask = make_causal_mask(seq_len) if use_explicit_mask else None
        output = encoder(x, mask=mask, is_causal=is_causal, inference=True)

        assert output.shape == (seq_len, d_model)

    def test_with_process_heads(self):
        d_model = 64
        nhead = 4
        num_layers = 3
        seq_len = 10

        encoder = TransformerEncoder(
            d_model=d_model,
            nhead=nhead,
            num_layers=num_layers,
            key=jax.random.key(0),
            dtype=jnp.float32,
        )

        def identity_process_heads(q, k, v):
            return q, k, v

        x = jax.random.normal(jax.random.key(1), (seq_len, d_model))
        output = encoder(x, process_heads=identity_process_heads, inference=True)

        assert output.shape == (seq_len, d_model)

    def test_jit_no_retrace(self):
        d_model = 64
        nhead = 4
        num_layers = 3
        seq_len = 10

        encoder = TransformerEncoder(
            d_model=d_model,
            nhead=nhead,
            num_layers=num_layers,
            key=jax.random.key(0),
            dtype=jnp.float32,
        )

        @eqx.filter_jit
        @eqx.debug.assert_max_traces(max_traces=1)
        def forward(model, x, inference):
            return model(x, inference=inference)

        x = jax.random.normal(jax.random.key(1), (seq_len, d_model))

        output1 = forward(encoder, x, True)
        output2 = forward(encoder, x, True)
        output3 = forward(encoder, x, True)

        assert output1.shape == (seq_len, d_model)
        assert output2.shape == (seq_len, d_model)
        assert output3.shape == (seq_len, d_model)


class TestTransformerDecoder:
    @pytest.mark.parametrize(
        "tgt_is_causal, memory_is_causal, use_tgt_mask, use_memory_mask",
        [
            (False, False, False, False),
            (True, False, False, False),
            (False, False, True, False),
            (False, False, False, True),
            (True, False, True, True),
        ],
    )
    def test_masking(
        self, tgt_is_causal, memory_is_causal, use_tgt_mask, use_memory_mask
    ):
        d_model = 64
        nhead = 4
        num_layers = 3
        tgt_len = 10
        src_len = 12

        decoder = TransformerDecoder(
            d_model=d_model,
            nhead=nhead,
            num_layers=num_layers,
            key=jax.random.key(0),
            dtype=jnp.float32,
        )

        tgt = jax.random.normal(jax.random.key(1), (tgt_len, d_model))
        memory = jax.random.normal(jax.random.key(2), (src_len, d_model))
        tgt_mask = make_causal_mask(tgt_len) if use_tgt_mask else None
        memory_mask = (
            jnp.ones((tgt_len, src_len), dtype=jnp.bool_) if use_memory_mask else None
        )

        output = decoder(
            tgt,
            memory,
            tgt_mask=tgt_mask,
            memory_mask=memory_mask,
            tgt_is_causal=tgt_is_causal,
            memory_is_causal=memory_is_causal,
            inference=True,
        )

        assert output.shape == (tgt_len, d_model)

    def test_with_process_heads(self):
        d_model = 64
        nhead = 4
        num_layers = 3
        tgt_len = 10
        src_len = 12

        decoder = TransformerDecoder(
            d_model=d_model,
            nhead=nhead,
            num_layers=num_layers,
            key=jax.random.key(0),
            dtype=jnp.float32,
        )

        def identity_process_heads(q, k, v):
            return q, k, v

        tgt = jax.random.normal(jax.random.key(1), (tgt_len, d_model))
        memory = jax.random.normal(jax.random.key(2), (src_len, d_model))
        output = decoder(
            tgt,
            memory,
            process_heads_sa=identity_process_heads,
            process_heads_mha=identity_process_heads,
            inference=True,
        )

        assert output.shape == (tgt_len, d_model)

    def test_jit_no_retrace(self):
        d_model = 64
        nhead = 4
        num_layers = 3
        tgt_len = 10
        src_len = 12

        decoder = TransformerDecoder(
            d_model=d_model,
            nhead=nhead,
            num_layers=num_layers,
            key=jax.random.key(0),
            dtype=jnp.float32,
        )

        @eqx.filter_jit
        @eqx.debug.assert_max_traces(max_traces=1)
        def forward(model, tgt, memory, inference):
            return model(tgt, memory, inference=inference)

        tgt = jax.random.normal(jax.random.key(1), (tgt_len, d_model))
        memory = jax.random.normal(jax.random.key(2), (src_len, d_model))

        output1 = forward(decoder, tgt, memory, True)
        output2 = forward(decoder, tgt, memory, True)
        output3 = forward(decoder, tgt, memory, True)

        assert output1.shape == (tgt_len, d_model)
        assert output2.shape == (tgt_len, d_model)
        assert output3.shape == (tgt_len, d_model)


class TestTransformer:
    @pytest.mark.parametrize(
        "src_is_causal, tgt_is_causal, memory_is_causal, use_src_mask, use_tgt_mask, use_memory_mask",
        [
            (False, False, False, False, False, False),
            (True, True, False, False, False, False),
            (False, False, False, True, True, True),
            (True, True, False, True, True, True),
        ],
    )
    def test_masking(
        self,
        src_is_causal,
        tgt_is_causal,
        memory_is_causal,
        use_src_mask,
        use_tgt_mask,
        use_memory_mask,
    ):
        d_model = 64
        nhead = 4
        src_len = 12
        tgt_len = 10

        transformer = Transformer(
            d_model=d_model,
            nhead=nhead,
            key=jax.random.key(0),
            dtype=jnp.float32,
        )

        src = jax.random.normal(jax.random.key(1), (src_len, d_model))
        tgt = jax.random.normal(jax.random.key(2), (tgt_len, d_model))
        src_mask = make_causal_mask(src_len) if use_src_mask else None
        tgt_mask = make_causal_mask(tgt_len) if use_tgt_mask else None
        memory_mask = (
            jnp.ones((tgt_len, src_len), dtype=jnp.bool_) if use_memory_mask else None
        )

        output = transformer(
            src,
            tgt,
            src_mask=src_mask,
            tgt_mask=tgt_mask,
            memory_mask=memory_mask,
            src_is_causal=src_is_causal,
            tgt_is_causal=tgt_is_causal,
            memory_is_causal=memory_is_causal,
            inference=True,
        )

        assert output.shape == (tgt_len, d_model)

    def test_with_process_heads(self):
        d_model = 64
        nhead = 4
        src_len = 12
        tgt_len = 10

        transformer = Transformer(
            d_model=d_model,
            nhead=nhead,
            key=jax.random.key(0),
            dtype=jnp.float32,
        )

        def identity_process_heads(q, k, v):
            return q, k, v

        src = jax.random.normal(jax.random.key(1), (src_len, d_model))
        tgt = jax.random.normal(jax.random.key(2), (tgt_len, d_model))
        output = transformer(
            src,
            tgt,
            process_heads_enc=identity_process_heads,
            process_heads_dec_sa=identity_process_heads,
            process_heads_dec_mha=identity_process_heads,
            inference=True,
        )

        assert output.shape == (tgt_len, d_model)

    @pytest.mark.parametrize(
        "activation",
        [
            jax.nn.relu,
            jax.nn.gelu,
            jax.nn.silu,
        ],
    )
    def test_activations(self, activation):
        d_model = 64
        nhead = 4
        src_len = 12
        tgt_len = 10

        transformer = Transformer(
            d_model=d_model,
            nhead=nhead,
            activation=activation,
            key=jax.random.key(0),
            dtype=jnp.float32,
        )

        src = jax.random.normal(jax.random.key(1), (src_len, d_model))
        tgt = jax.random.normal(jax.random.key(2), (tgt_len, d_model))
        output = transformer(src, tgt, inference=True)

        assert output.shape == (tgt_len, d_model)

    def test_jit_no_retrace(self):
        d_model = 64
        nhead = 4
        src_len = 12
        tgt_len = 10

        transformer = Transformer(
            d_model=d_model,
            nhead=nhead,
            key=jax.random.key(0),
            dtype=jnp.float32,
        )

        @eqx.filter_jit
        @eqx.debug.assert_max_traces(max_traces=1)
        def forward(model, src, tgt, inference):
            return model(src, tgt, inference=inference)

        src = jax.random.normal(jax.random.key(1), (src_len, d_model))
        tgt = jax.random.normal(jax.random.key(2), (tgt_len, d_model))

        output1 = forward(transformer, src, tgt, True)
        output2 = forward(transformer, src, tgt, True)
        output3 = forward(transformer, src, tgt, True)

        assert output1.shape == (tgt_len, d_model)
        assert output2.shape == (tgt_len, d_model)
        assert output3.shape == (tgt_len, d_model)



import abc

import equinox as eqx
from jaxtyping import Array


class AbstractNorm(eqx.Module):
    @abc.abstractmethod
    def __call__(self, x: Array, *_, **__) -> Array: ...


class AbstractNormStateful(eqx.nn.StatefulLayer):
    @abc.abstractmethod
    def __call__(
        self, x: Array, state: eqx.nn.State, *_, **__
    ) -> tuple[Array, eqx.nn.State]: ...



import equinox as eqx
import jax
import jax.numpy as jnp
from beartype.typing import Any, Hashable, Sequence
from equinox.nn import State
from jaxtyping import Array, Float, PRNGKeyArray

from jaxonlayers.functions.utils import default_floating_dtype

from .abstract import AbstractNorm, AbstractNormStateful


class BatchNorm(AbstractNormStateful):
    state_index: eqx.nn.StateIndex

    gamma: Float[Array, "size"] | None
    beta: Float[Array, "size"] | None

    inference: bool
    axis_name: Hashable | Sequence[Hashable]

    size: int = eqx.field(static=True)
    eps: float = eqx.field(static=True)
    momentum: float = eqx.field(static=True)
    affine: bool = eqx.field(static=True)

    def __init__(
        self,
        size: int,
        axis_name: Hashable | Sequence[Hashable],
        eps: float = 1e-5,
        momentum: float = 0.1,
        affine: bool = True,
        inference: bool = False,
        dtype: Any | None = None,
    ):
        if dtype is None:
            dtype = default_floating_dtype()
        assert dtype is not None
        self.size = size
        self.eps = eps
        self.momentum = momentum
        self.affine = affine
        self.inference = inference
        self.axis_name = axis_name

        self.gamma = jnp.ones(self.size, dtype=dtype) if self.affine else None
        self.beta = jnp.zeros(self.size, dtype=dtype) if self.affine else None

        self.state_index = eqx.nn.StateIndex(
            (jnp.zeros(size, dtype=dtype), jnp.ones(size, dtype=dtype))
        )

    def __call__(
        self,
        x: Array,
        state: State,
        *,
        key: PRNGKeyArray | None = None,
    ) -> tuple[Array, State]:
        running_mean, running_var = state.get(self.state_index)

        input_shape = x.shape
        ndim = len(input_shape)

        if ndim == 1:
            batch_mean = jax.lax.pmean(x, axis_name=self.axis_name)
            batch_size = jax.lax.psum(1, axis_name=self.axis_name)

            if self.inference:
                x_normalized = (x - running_mean) / jnp.sqrt(running_var + self.eps)
            else:
                xmu = x - batch_mean
                sq = xmu**2
                batch_var = jax.lax.pmean(sq, axis_name=self.axis_name)
                std = jnp.sqrt(batch_var + self.eps)
                x_normalized = xmu / std

                correction_factor = batch_size / jnp.maximum(batch_size - 1, 1)
                running_mean = (
                    1 - self.momentum
                ) * running_mean + self.momentum * batch_mean
                running_var = (1 - self.momentum) * running_var + self.momentum * (
                    batch_var * correction_factor
                )

                state = state.set(self.state_index, (running_mean, running_var))
        else:
            spatial_axes = tuple(range(1, ndim))  # All dims except channel dim (0)

            if self.inference:
                x_normalized = (
                    x - running_mean.reshape((-1,) + (1,) * (ndim - 1))
                ) / jnp.sqrt(running_var.reshape((-1,) + (1,) * (ndim - 1)) + self.eps)
            else:
                spatial_mean = jnp.mean(x, axis=spatial_axes)

                batch_mean = jax.lax.pmean(spatial_mean, axis_name=self.axis_name)
                batch_size = jax.lax.psum(1, axis_name=self.axis_name)

                broadcast_shape = (-1,) + (1,) * (ndim - 1)
                batch_mean_broadcasted = batch_mean.reshape(broadcast_shape)

                xmu = x - batch_mean_broadcasted
                sq = xmu**2

                spatial_var = jnp.mean(sq, axis=spatial_axes)
                batch_var = jax.lax.pmean(spatial_var, axis_name=self.axis_name)

                batch_var_broadcasted = batch_var.reshape(broadcast_shape)
                std = jnp.sqrt(batch_var_broadcasted + self.eps)

                x_normalized = xmu / std

                spatial_size = 1
                for dim in spatial_axes:
                    spatial_size *= x.shape[dim]
                total_size = batch_size * spatial_size

                correction_factor = total_size / jnp.maximum(total_size - 1, 1)
                running_mean = (
                    1 - self.momentum
                ) * running_mean + self.momentum * batch_mean
                running_var = (1 - self.momentum) * running_var + self.momentum * (
                    batch_var * correction_factor
                )

                state = state.set(self.state_index, (running_mean, running_var))

        out = x_normalized
        if self.affine and self.gamma is not None and self.beta is not None:
            if ndim > 1:
                broadcast_shape = (-1,) + (1,) * (ndim - 1)
                gamma_broadcasted = self.gamma.reshape(broadcast_shape)
                beta_broadcasted = self.beta.reshape(broadcast_shape)
                out = gamma_broadcasted * x_normalized + beta_broadcasted
            else:
                out = self.gamma * x_normalized + self.beta

        return out, state


class LocalResponseNormalization(eqx.Module):
    k: int = eqx.field(static=True)
    n: int = eqx.field(static=True)
    alpha: float = eqx.field(static=True)
    beta: float = eqx.field(static=True)

    def __init__(self, k=2, n=5, alpha=1e-4, beta=0.75) -> None:
        self.k = k
        self.n = n
        self.alpha = alpha
        self.beta = beta

    def __call__(self, x: Float[Array, "c h w"]) -> Float[Array, "c h w"]:
        c, _, _ = x.shape
        p = jnp.pad(x, pad_width=[(self.n // 2, self.n // 2), (0, 0), (0, 0)])

        def _body(i):
            window = jax.lax.dynamic_slice_in_dim(p, i, self.n) ** 2
            d = (jnp.einsum("ijk->jk", window) * self.alpha + self.k) ** self.beta
            b = x[i] / d
            return b

        ys = eqx.filter_vmap(_body)(jnp.arange(c))
        return ys


class LayerNorm(AbstractNorm):
    shape: tuple[int, ...] = eqx.field(static=True)
    axes: tuple[int, ...] = eqx.field(static=True)
    eps: float = eqx.field(static=True)
    use_weight: bool = eqx.field(static=True)
    use_bias: bool = eqx.field(static=True)
    weight: Array | None
    bias: Array | None

    def __init__(
        self,
        shape: int | Sequence[int],
        eps: float = 1e-5,
        use_weight: bool = True,
        use_bias: bool = True,
        dtype=None,
    ):
        if isinstance(shape, int):
            shape_tuple = (shape,)
        else:
            shape_tuple = tuple(shape)
        self.shape = shape_tuple
        self.axes = tuple(range(-len(self.shape), 0))
        self.eps = eps
        self.use_weight = use_weight
        self.use_bias = use_bias

        if dtype is None:
            dtype = default_floating_dtype()
        assert dtype is not None

        if self.use_weight:
            self.weight = jnp.ones(self.shape, dtype=dtype)
        else:
            self.weight = None
        if self.use_bias:
            self.bias = jnp.zeros(self.shape, dtype=dtype)
        else:
            self.bias = None

    def __call__(
        self,
        x: Array,
        *,
        key: PRNGKeyArray | None = None,
    ) -> Array:
        if x.shape[-len(self.shape) :] != self.shape:
            raise ValueError(f"Input shape {x.shape} must end with shape {self.shape}")

        orig_dtype = x.dtype
        with jax.numpy_dtype_promotion("standard"):
            calc_dtype_elems = [x.dtype, jnp.float32]
            if self.weight is not None:
                calc_dtype_elems.append(self.weight.dtype)
            if self.bias is not None:
                calc_dtype_elems.append(self.bias.dtype)
            calc_dtype = jnp.result_type(*calc_dtype_elems)

        x = x.astype(calc_dtype)

        mean = jnp.mean(x, axis=self.axes, keepdims=True)
        mean_keepdims = jnp.mean(x, axis=self.axes, keepdims=True)
        variance = jnp.mean(
            jnp.square(x - mean_keepdims), axis=self.axes, keepdims=True
        )
        variance = jnp.maximum(0.0, variance)
        inv = jax.lax.rsqrt(variance + self.eps)
        out = (x - mean) * inv

        if self.use_weight:
            assert self.weight is not None
            out = out * self.weight.astype(calc_dtype)
        if self.use_bias:
            assert self.bias is not None
            out = out + self.bias.astype(calc_dtype)

        out = out.astype(orig_dtype)
        return out



from itertools import repeat

import equinox as eqx
import jax
from beartype.typing import Any, Callable, Sequence
from equinox.nn import StatefulLayer
from jaxtyping import Array, Float, PRNGKeyArray

from jaxonlayers.layers.abstract import AbstractNorm, AbstractNormStateful


def make_ntuple(x: Any, n: int) -> tuple[Any, ...]:
    if isinstance(x, collections.abc.Iterable):  # pyright: ignore
        return tuple(x)
    return tuple(repeat(x, n))


class ConvNormActivation(StatefulLayer):
    conv: eqx.nn.Conv
    norm: AbstractNorm | AbstractNormStateful | None
    activation: eqx.nn.Lambda | None

    def __init__(
        self,
        num_spatial_dims: int,
        in_channels: int,
        out_channels: int,
        kernel_size: int | Sequence[int] = 3,
        stride: int | Sequence[int] = 1,
        padding: str | int | Sequence[int] | Sequence[tuple[int, int]] | None = None,
        groups: int = 1,
        activation_layer: Callable[..., Array] | None = jax.nn.relu,
        dilation: int | Sequence[int] = 1,
        use_bias: bool | None = None,
        *,
        norm_layer: Callable[..., AbstractNorm | AbstractNormStateful] | None,
        key: PRNGKeyArray,
        dtype: Any,
    ):
        if padding is None:
            if isinstance(kernel_size, int) and isinstance(dilation, int):
                padding = (kernel_size - 1) // 2 * dilation
            else:
                _conv_dim = (
                    len(kernel_size)
                    if isinstance(kernel_size, Sequence)
                    else len(dilation)  # pyright: ignore
                )
                kernel_size = make_ntuple(kernel_size, _conv_dim)
                dilation = make_ntuple(dilation, _conv_dim)
                padding = tuple(
                    (kernel_size[i] - 1) // 2 * dilation[i] for i in range(_conv_dim)
                )
        if use_bias is None:
            use_bias = norm_layer is None

        key, subkey = jax.random.split(key)

        self.conv = eqx.nn.Conv(
            num_spatial_dims=num_spatial_dims,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            use_bias=use_bias,
            dtype=dtype,
            key=subkey,
        )

        self.norm = None
        if norm_layer is not None:
            self.norm = norm_layer()

        if activation_layer is not None:
            self.activation = eqx.nn.Lambda(activation_layer)
        else:
            self.activation = None

    def __call__(
        self,
        x: Float[Array, "c *num_spatial_dims"],
        state: eqx.nn.State,
        key: PRNGKeyArray | None = None,
    ) -> tuple[Float[Array, "c_out *num_spatial_dims_out"], eqx.nn.State]:
        x = self.conv(x)

        if self.norm:
            if isinstance(self.norm, StatefulLayer) and self.norm.is_stateful():
                assert state is not None
                x, state = self.norm(x, state)  # pyright: ignore
            else:
                x = self.norm(x)  # pyright: ignore

        if self.activation:
            x = self.activation(x)

        return x, state



import equinox as eqx
import jax
import jax.numpy as jnp
from beartype.typing import Any
from jaxtyping import Array, Float, PRNGKeyArray

from jaxonlayers.functions import default_floating_dtype, selective_scan


class SelectiveStateSpace(eqx.Module):
    input_proj: eqx.nn.Linear
    delta_proj: eqx.nn.Linear
    A_log: Float[Array, "d_inner d_state"]
    D: Float[Array, "d_inner"]
    out_proj: eqx.nn.Linear

    d_inner: int = eqx.field(static=True)
    dt_rank: int = eqx.field(static=True)
    d_state: int = eqx.field(static=True)
    d_model: int = eqx.field(static=True)

    def __init__(
        self,
        d_model: int,
        d_inner: int,
        dt_rank: int,
        d_state: int,
        use_input_proj_bias: bool = False,
        use_delta_proj_bias: bool = False,
        *,
        key: PRNGKeyArray,
        dtype: Any | None = None,
    ):
        if dtype is None:
            dtype = default_floating_dtype()
        assert dtype is not None

        self.d_model = d_model
        self.d_inner = d_inner
        self.dt_rank = dt_rank
        self.d_state = d_state

        keys = jax.random.split(key, 4)
        proj_dim = self.dt_rank + 2 * self.d_inner * self.d_state
        self.input_proj = eqx.nn.Linear(
            self.d_model,
            proj_dim,
            use_bias=use_input_proj_bias,
            key=keys[0],
            dtype=dtype,
        )

        self.delta_proj = eqx.nn.Linear(
            dt_rank,
            d_inner,
            use_bias=use_delta_proj_bias,
            key=keys[1],
            dtype=dtype,
        )

        A = jnp.arange(1, d_state + 1, dtype=jnp.float32)
        A = jnp.tile(A, (d_inner, 1))
        self.A_log = jnp.log(A)

        self.D = jnp.ones(d_inner, dtype=dtype)

        self.out_proj = eqx.nn.Linear(
            d_inner, d_model, use_bias=False, key=keys[2], dtype=dtype
        )

    def __call__(self, x: Float[Array, "seq_length d_inner"]):
        L, _ = x.shape
        A = -jnp.exp(self.A_log.astype(jnp.float32))
        D = self.D.astype(jnp.float32)

        delta_b_c = jax.vmap(self.input_proj)(x)
        delta, B, C = jnp.split(
            delta_b_c,
            [self.dt_rank, self.dt_rank + self.d_inner * self.d_state],
            axis=-1,
        )

        B = B.reshape(L, self.d_inner, self.d_state)
        C = C.reshape(L, self.d_inner, self.d_state)

        delta = jax.nn.softplus(jax.vmap(self.delta_proj)(delta))

        y = selective_scan(x, delta, A, B, C, D)

        return jax.vmap(self.out_proj)(y)
